import numpy as np
import pygame
import sys
import matplotlib

from matplotlib import pyplot as plt

from particle_generator import ParticleGenerator
from ui.button import Button
from particle import Particle
from rectangle import Rectangle
from ui.input_box import InputBox
from ui.drop_menu import DropDownMenu
from scipy.interpolate import make_interp_spline

matplotlib.use('module://pygame_matplotlib.backend_pygame')


class DemoScreen:
    def __init__(self, app, window_size):
        self.app = app
        self.screen = app.screen
        self.speed = 1
        self.bg_color = (255, 255, 255)
        self.font = 'arial'
        self.width, self.length = window_size
        self.curr_tick = 0
        self.pause = False
        self.dt = 0
        self.left_particles_counter = 0
        self.right_particles_counter = 0

        self.little_font = pygame.font.SysFont(self.font, self.length // 40)
        self.middle_font_bold = pygame.font.SysFont(self.font, self.length // 27, bold=True)
        self.middle_font = pygame.font.SysFont(self.font, self.length // 27)

        self.graphics_names = ['Смещение', 'Скорость', 'Ускорение', 'Кинетическая энергия']

        self.graph_size = (self.width // 4, self.length // 8)

        fig_size = (self.width / 180, self.length / 125)

        self.figure, self.axis = plt.subplots(4, 1, figsize=fig_size, sharex=True)

        self.fig_size = self.figure.get_size_inches() * self.figure.dpi
        self.fig_start = 0

        plt.subplots_adjust(left=0.12,
                            bottom=0.1,
                            right=0.95,
                            top=0.95,
                            wspace=0.2,
                            hspace=0.4)

        matplotlib.rcParams.update({'font.size': 16})

        # Границы движения
        self.ox_start = self.width // 48
        self.ox_end = 47 * self.width // 48
        self.ox_len = self.ox_end - self.ox_start
        self.ox_zero = self.width // 2

        self.ox_bottom = 37 * self.length // 40
        self.ox_top = 32 * self.length // 40
        self.ox_ruler = 38 * self.length // 40

        # Лайаут для настроек параметров
        self.option_surface_margin_width = self.width // 10
        self.option_surface_margin_height = self.width // 30

        self.option_surface_width = self.width // 2 - self.option_surface_margin_width
        self.option_surface_length = 3 * self.fig_size[1] / 4

        self.option_surface_margin_in_surface = self.option_surface_length / 20

        self.option_surface_size = (self.option_surface_width,
                                    self.option_surface_length - self.option_surface_margin_height)
        self.option_surface_coords = (self.fig_size[0], self.option_surface_margin_height)

        self.option_surface = pygame.Surface(self.option_surface_size)

        # Кнопки
        button_size = (self.width // 8, self.length // 16)
        margin_between_buttons = (self.option_surface_width - 3 * button_size[0]) / 2
        button_pos3 = (self.option_surface_coords[0],
                       self.option_surface_length + self.option_surface_margin_height)

        button_pos1 = (button_pos3[0] + button_size[0] + margin_between_buttons,
                       self.option_surface_length + self.option_surface_margin_height)

        button_pos2 = (button_pos1[0] + button_size[0] + margin_between_buttons,
                       self.option_surface_length + self.option_surface_margin_height)

        self.buttons = [Button(app, 'Заново', button_pos1, button_size),
                        Button(app, 'Пауза', button_pos3, button_size),
                        Button(app, "Назад", button_pos2, button_size)]

        # Создание объектов внутри option_surface
        box_input_size = (self.option_surface_size[0] // 5, self.option_surface_size[1] // 10)
        self.input_boxes = []
        input_boxes_names = ['M', 'm', 'T', 'n', 'p слева', 'период']
        types = [float, float, int, int, float, int]
        input_boxes_default_values = [1., 0.01, 300, 100, 0.5, 4]
        self.box_values = {}
        margin_between_input_boxes = (self.option_surface_width - 3 * box_input_size[0]) // 5
        for i, name in enumerate(input_boxes_names):
            self.input_boxes.append(
                InputBox((margin_between_input_boxes + (i % 2) * (margin_between_input_boxes + box_input_size[0]),
                          (i // 2 + 1) * 2 * self.option_surface_length // 10),
                         box_input_size, input_boxes_default_values[i], name, types[i]))
            self.box_values[name] = input_boxes_default_values[i]

        drop_menu_size = (
        2 * self.option_surface_size[0] // 5 - margin_between_input_boxes, self.option_surface_size[1] // 10)

        self.modes = ['Периодический', 'Случайный', 'Квазипериодич.', 'Квазисимметр.', 'Cимметричный']
        self.drop_menu = DropDownMenu(
            (margin_between_input_boxes + 2 * (margin_between_input_boxes + box_input_size[0]),
             2 * self.option_surface_length // 10), drop_menu_size,
            self.modes, 'Режим')

        self.left_particles_coords = (margin_between_input_boxes, 29 * self.option_surface_length // 40)
        self.right_particles_coords = (margin_between_input_boxes, 32 * self.option_surface_length // 40)

        # Генерация объектов
        self.rectangle = Rectangle(self.width // 2 - self.ox_len // 20, self.ox_top + 1,
                                   self.ox_len // 10, self.ox_bottom - self.ox_top - 1, 1)

        self.particles = []
        self.particles_y = (self.ox_top + self.ox_bottom) // 2

        self._start_new_lap()

    def _start_new_lap(self):

        self.T = self.input_boxes[2].value if 100 <= self.input_boxes[2].value <= 500 else 300
        self.p = self.input_boxes[4].value if 0 <= self.input_boxes[4].value <= 1 else 0.5

        self.m = self.input_boxes[1].value if 0.0001 <= self.input_boxes[1].value <= 1 else 0.01
        self.M = self.input_boxes[0].value if 0.01 <= self.input_boxes[0].value <= 10 else 1
        self.n = int(self.input_boxes[3].value) if 1 <= int(self.input_boxes[3].value) <= 100 else 100
        self.t = self.input_boxes[5].value if 1 <= int(self.input_boxes[5].value) <= 40 else 4

        self.particle_generator = ParticleGenerator(self.modes.index(self.drop_menu.cur_option), self.t, self.p,
                                                    self.ox_start,
                                                    self.ox_end, self.particles_y, self.m, self.length // 160, self.T,
                                                    self.rectangle.x, self.rectangle.x + self.rectangle.width)

        params = [self.M, self.m, self.T, self.n, self.p, self.t]

        for i, box in enumerate(self.input_boxes):
            box.set_value(params[i])
            box.draw(self.option_surface)

        self.rectangle.x = self.width // 2 - self.rectangle.width // 2
        self.rectangle.v_x = 0
        self.rectangle.a_x = 0
        self.particles = []

        # Генерация данных для графиков
        self.x = [0]
        self.rect_speeds = [0]
        self.rect_coords = [0]
        self.rect_kinetic = [0]
        self.rect_acceleration = [0]
        self.data = [self.rect_coords, self.rect_speeds, self.rect_acceleration, self.rect_kinetic]

        for i, y in enumerate(self.data):
            self.axis[i].clear()

        self.left_particles_counter = 0
        self.right_particles_counter = 0
        self.curr_tick = 0
        self.pause = False

    def _update_screen(self):
        self.curr_tick = (self.curr_tick + 1) % self.t
        self.screen.fill(self.bg_color)

        for button in self.buttons:
            button.draw_button()

        self.option_surface.fill((245, 245, 245))

        pygame.draw.rect(self.option_surface, (0, 0, 0), (0, 0, self.option_surface_width,
                                                          self.option_surface_length - self.option_surface_margin_height),
                         1,
                         border_radius=10)
        option_name = self.middle_font.render('Настройка параметров:', True, (0, 0, 0))

        left_particles = self.middle_font.render(f'Лево: {self.left_particles_counter}', False, (0, 0, 0))
        right_particles = self.middle_font.render(f'Право: {self.right_particles_counter}', False, (0, 0, 0))

        self.option_surface.blit(right_particles, self.left_particles_coords)
        self.option_surface.blit(left_particles, self.right_particles_coords)

        self.option_surface.blit(option_name, (self.option_surface_width // 60, self.option_surface_length // 40))

        for box in self.input_boxes:
            box.draw(self.option_surface)

        self.drop_menu.draw(self.option_surface)
        self.screen.blit(self.option_surface, self.option_surface_coords)

        # Границы движения
        pygame.draw.line(self.screen, (0, 0, 0), (self.ox_start, self.ox_bottom),
                         (self.ox_end, self.ox_bottom))

        pygame.draw.line(self.screen, (0, 0, 0), (self.ox_start, self.ox_top),
                         (self.ox_end, self.ox_top))

        # Графики
        self._draw_graphics()
        self._draw_ruler()

        if self.rectangle.x + self.rectangle.width >= self.ox_end or self.rectangle.x <= self.ox_start:
            self.pause = True

        if self.pause:
            self._draw_particles(False)
            self.rectangle.show(self.screen)
            return

        self._draw_particles(True)
        self.rectangle.move()
        self.rectangle.show(self.screen)

        self.rect_speeds.append(self.rectangle.v_x)
        self.rect_coords.append(self.rectangle.x_center - self.ox_zero)
        self.rect_kinetic.append(self.rectangle.m * self.rectangle.v_x ** 2 / 2)
        self.rect_acceleration.append(self.rectangle.a_x)
        self.x.append(self.x[-1] + 1)
        self.data = [self.rect_coords, self.rect_speeds, self.rect_acceleration, self.rect_kinetic]

        self.rectangle.a_x = 0

    def _draw_graphics(self):
        for i, y in enumerate(self.data):
            axes = self.axis[i]
            axes.clear()
            axes.plot(self.x, np.mean(y).repeat(len(self.x)), color='red')

            if i == 3 and len(self.x) > 2:
                x_new = np.linspace(self.x[0], self.x[-1], self.x[-1] * 4)
                spl = make_interp_spline(self.x, y, k=2)
                power_smooth = spl(x_new)
                axes.plot(x_new, power_smooth, color='black')
            else:
                axes.plot(self.x, y, color='black')

            axes.set_title(self.graphics_names[i])

            axes.get_figure().canvas.draw()
        self.screen.blit(self.figure, (self.fig_start, 0))

    def _draw_particles(self, move):
        if not self.pause:
            self.dt += 1

            new_particle = self.particle_generator.get_particle(self.curr_tick)

            if not new_particle is None and len(self.particles) < self.n:
                if len(new_particle) == 2:
                    self.left_particles_counter += 1
                    self.right_particles_counter += 1
                elif new_particle[0].x == self.ox_start:
                    self.left_particles_counter += 1
                else:
                    self.right_particles_counter += 1

                self.particles.extend(new_particle)

        was_hit = False
        remove_list = []
        prev_rect_v = self.rectangle.v_x
        for i,particle in enumerate(self.particles):
            if move:
                particle.move()
            if ((self.rectangle.x <= particle.x and particle.v_x > 0) or
                    (particle.x <= self.rectangle.x + self.rectangle.width) and particle.v_x < 0):
                was_hit = True
                if particle.v_x < 0:
                    Particle(x=self.rectangle.x + self.rectangle.width, y=self.particles_y,
                             v_x=0, r=self.length // 120, color=(244, 169, 0)).show(self.screen)
                else:
                    Particle(x=self.rectangle.x, y=self.particles_y,
                             v_x=0, r=self.length // 120, color=(244, 169, 0)).show(self.screen)

                self.rectangle.v_x = (2 * self.m * particle.v_x + self.rectangle.v_x * (self.M - self.m)) / (self.m + self.M)
                remove_list.append(particle)
            else:
                particle.show(self.screen)

        if was_hit:
            for removing in remove_list:
                self.particles.remove(removing)
            self.rectangle.a_x = (self.rectangle.v_x - prev_rect_v) / self.dt
            self.dt = 1


    def _check_events(self):
        for event in pygame.event.get():
            if event.type == pygame.QUIT:
                sys.exit()
            for box in self.input_boxes:
                box.handle_event(event, self.option_surface_coords)

            self.drop_menu.handle_event(event, self.option_surface_coords)

            if event.type == pygame.MOUSEBUTTONDOWN:
                mouse_position = pygame.mouse.get_pos()
                self._check_buttons(mouse_position)

    def _check_buttons(self, mouse_position):
        for index, button in enumerate(self.buttons):
            if button.rect.collidepoint(mouse_position):
                button.pressed = True
                if index == 0:
                    self._start_new_lap()
                if index == 1:
                    self.pause = not self.pause
                if index == 2:
                    self.app.active_screen = self.app.menu_screen

    def _draw_ruler(self):
        i = 0
        pygame.draw.line(self.screen, (0, 0, 0), (self.ox_start, self.ox_ruler), (self.ox_end, self.ox_ruler))

        start = self.ox_start - self.ox_start % 10
        for x in range(start, self.ox_end + 1, 5):
            if (x - self.ox_zero) % 200 == 0:
                pygame.draw.line(self.screen, (255, 0, 0), (x, self.ox_ruler), (x, self.ox_ruler - 10), 1)
                self._ruler_labels(x, self.ox_ruler, '{0}'.format(x - self.ox_zero))
            else:
                pygame.draw.line(self.screen, (255, 0, 0), (x, self.ox_ruler), (x, self.ox_ruler - 5), 1)
            i += 1

    def _ruler_labels(self, x, y, t):
        text = self.little_font.render(t, True, (255, 0, 0))
        r = text.get_rect()
        self.screen.blit(text, (x - r.width / 2, y + 1))
